var topic = "<!-- saved from url=(0024)http://docs.autodesk.com -->\n\
<html>\n\
   <head>\n\
<link href=\"../../style/prettify.css\" type=\"text/css\" rel=\"stylesheet\" />\n\
<script type=\"text/javascript\" src=\"../../scripts/prettify.js\"></script>\n\
<script type=\"text/javascript\" src=\"../../scripts/lib/jquery-1.11.1.min.js\"></script><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"><meta http-equiv=\"Content-Style-Type\" content=\"text/css\"><meta name=\"generator\" content=\"pandoc\"><meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\"><script type=\"text/javascript\" src=\"../../scripts/utils/adsk.redirect.js\"></script>\n\
      <title>Threading and Maya API</title>\n\
   </head>\n\
   <body height=\"100%\"><div class=\"body_content\" id=\"body-content\"><style type=\"text/css\">code{white-space: pre;}</style><script>$(document).ready(function () { prettyPrint(); } );</script>\n\
      <div>\n\
         <div class=\"head\">\n\
            <h1>Threading and Maya API</h1>\n\
         </div>\n\
\n\
<div class=\'section\'><a id=\"threading-and-maya-api\"></a></div>\n\
<p>Plug-ins can utilize threaded code if certain guidelines are followed. Below are some notes that provide guidance for this issue:</p>\n\
<ol>\n\
<li>Maya uses the following types of threads:\n\
<ul>\n\
<li>Linux - pthreads</li>\n\
<li>Mac OS X - pthreads</li>\n\
<li>Windows - Native Windows API threads</li>\n\
</ul></li>\n\
<li>The components of Maya that are available in the API are single threaded. It is always best to call into the Maya API from the main Maya thread. It is acceptable to thread your code as long as it is independent of calls to the Maya API. The exception to this rule is <span class=\'code\'>MPxNode::compute()</span> can have threaded calls for software shaders. But this depends on the setup of the node and the software renderer.</li>\n\
<li>Although some operations in Maya have been threaded, they are not exposed in the API.</li>\n\
<li>It is possible to call into Maya from an secondary thread using the <span class=\'code\'>MGlobal::executeCommandOnIdle()</span> method. In Python, the equivalent <span class=\'code\'>MGlobal::executePythonCommandOnIdle()</span> method would be used. The command will not execute immediately; instead, the command will be added to the idle event queue and executed as idle processing allows. The result of the command will not be returned to the caller. This call can be useful for updating items such as the progress bar from another thread.</li>\n\
<li><p>There are four C++ API classes for threading:</p>\n\
<ul>\n\
<li><span class=\'code\'>MThreadPool</span></li>\n\
<li><span class=\'code\'>MThreadAsync</span></li>\n\
<li><span class=\'code\'>MSpinLock</span></li>\n\
<li><span class=\'code\'>MMutexLock</span></li>\n\
</ul>\n\
<p>These classes can be used to implement threaded algorithms for non-Maya API functionality. Several examples that utilize these classes can be found in the developer kit. (These classes are not available in the Maya Python API.)</p>\n\
<p><strong>MThreadPool</strong> gives access to a pool of threads to which tasks can be assigned. The number of tasks does not have to equal the number of threads, in fact for load balancing it is usually better if the number of tasks exceeds the number of threads. Maya will internally balance the work among the threads for optimal efficiency. The number of threads in the pool is equal to the number of logical processors. It is not necessary to delete the thread pool after each usage, and for performance reasons it is better not to do so, since the threads will be put to sleep when a parallel region finishes, which means they can be restarted quickly.</p>\n\
<p><strong>MThreadAsync</strong> allows the creation of one of more threads that can run for a long time. They are not drawn from the thread pool created and managed by <span class=\'code\'>MThreadPool</span>, but are independent threads. These threads can be used for longer running tasks. Since they are not created from the thread pool, the number and workload of such threads should be managed carefully to avoid oversubscription issues, where the number of busy threads exceeds the hardware resources available.</p>\n\
<p><strong>MMutexLock</strong> is a locking primitive that can be used with both <span class=\'code\'>MThreadPool</span> and <span class=\'code\'>MThreadAsync</span> threads. It allows standard mutex locking of threads.</p>\n\
<p><strong>MSpinLock</strong> is a lock that spin-waits, so can be more efficient than a mutex lock in situations where the lock is likely to be held for a very short time. However since the lock spin waits, it is a heavy CPU consumer, and should not be used when locks are likely to be held for a long time.</p></li>\n\
<li><p>Threading with Python is possible with the built-in thread module. The thread module can be used to implement threaded algorithms for non-Maya API functionality. Please see the Python and threading section of the <em>Python</em> Guide for more details.</p></li>\n\
</ol>\n\
<p>The following example demonstrates how to find primes using a serial and a threaded approach. The threaded approach uses the <span class=\'code\'>MThreadPool</span> class.</p>\n\
<div class=\"codeBlock\"><pre class=\"prettyprint\">#include &lt;math.h&gt;\n\
#include &lt;maya/MIOStream.h&gt;\n\
#include &lt;maya/MSimple.h&gt;\n\
#include &lt;maya/MTimer.h&gt;\n\
#include &lt;maya/MGlobal.h&gt;\n\
#include &lt;maya/MThreadPool.h&gt;\n\
DeclareSimpleCommand( threadTestCmd, PLUGIN_COMPANY, &quot;2017&quot;);\n\
typedef struct _threadDataTag\n\
{\n\
    int threadNo;\n\
    long primesFound;\n\
    long start, end;\n\
} threadData;\n\
\n\
typedef struct _taskDataTag\n\
{\n\
    long start, end, totalPrimes;\n\
} taskData;\n\
\n\
#define NUM_TASKS   16\n\
// No global information used in function\n\
\n\
static bool TestForPrime(int val)\n\
{\n\
    int limit, factor = 3;\n\
    limit = (long)(sqrtf((float)val)+0.5f);\n\
    while( (factor &lt;= limit) &amp;&amp; (val % factor))\n\
        factor ++;\n\
    return (factor &gt; limit);\n\
}\n\
\n\
// Primes finder. This function is called from multiple threads\n\
MThreadRetVal Primes(void *data)\n\
{\n\
    threadData *myData = (threadData *)data;\n\
    for( int i = myData-&gt;start + myData-&gt;threadNo*2; i &lt;= myData-&gt;end; i += 2*NUM_TASKS )\n\
    {\n\
        if( TestForPrime(i) )\n\
        myData-&gt;primesFound++;\n\
    }\n\
    return (MThreadRetVal)0;\n\
}\n\
\n\
// Function to create thread tasks\n\
void DecomposePrimes(void *data, MThreadRootTask *root)\n\
{\n\
    taskData *taskD = (taskData *)data;\n\
    threadData tdata[NUM_TASKS];\n\
    for( int i = 0; i &lt; NUM_TASKS; ++i )\n\
    {\n\
        tdata[i].threadNo = i;\n\
        tdata[i].primesFound = 0;\n\
        tdata[i].start = taskD-&gt;start;\n\
        tdata[i].end = taskD-&gt;end;\n\
        MThreadPool::createTask(Primes, (void *)&amp;tdata[i], root);\n\
    }\n\
    MThreadPool::executeAndJoin(root);\n\
    for( int i = 0; i &lt; NUM_TASKS; ++i )\n\
    {\n\
        taskD-&gt;totalPrimes += tdata[i].primesFound;\n\
    }\n\
}\n\
\n\
// Single threaded calculation\n\
int SerialPrimes(int start, int end)\n\
{\n\
    int primesFound = 0;\n\
    for( int i = start; i &lt;= end; i+=2)\n\
    {\n\
        if( TestForPrime(i) )\n\
            primesFound++;\n\
    }\n\
    return primesFound;\n\
}\n\
\n\
// Set up and tear down parallel tasks\n\
int ParallelPrimes(int start, int end)\n\
{\n\
    MStatus stat = MThreadPool::init();\n\
    if( MStatus::kSuccess != stat ) {\n\
        MString str = MString(&quot;Error creating threadpool&quot;);\n\
        MGlobal::displayError(str);\n\
        return 0;\n\
    }\n\
\n\
    taskData tdata;\n\
    tdata.totalPrimes = 0;\n\
    tdata.start = start;\n\
    tdata.end = end;\n\
    MThreadPool::newParallelRegion(DecomposePrimes, (void *)&amp;tdata);\n\
    // pool is reference counted. Release reference to current thread instance\n\
    MThreadPool::release();\n\
    // release reference to whole pool which deletes all threads\n\
    MThreadPool::release();\n\
    return tdata.totalPrimes;\n\
}\n\
\n\
// MSimple command that invokes the serial and parallel thread calculations\n\
MStatus threadTestCmd::doIt( const MArgList&amp; args )\n\
{\n\
    MString introStr = MString(&quot;Computation of primes using the Maya API&quot;);\n\
    MGlobal::displayInfo(introStr);\n\
    if(args.length() != 2) {\n\
        MString str = MString(&quot;Invalid number of arguments, usage: threadTestCmd 1 10000&quot;);\n\
        MGlobal::displayError(str);\n\
        return MStatus::kFailure;\n\
    }\n\
    MStatus stat;\n\
    int start = args.asInt( 0, &amp;stat );\n\
    if ( MS::kSuccess != stat ) {\n\
        MString str = MString(&quot;Invalid argument 1, usage: threadTestCmd 1 10000&quot;);\n\
        MGlobal::displayError(str);\n\
        return MStatus::kFailure;\n\
    }\n\
    int end = args.asInt( 1, &amp;stat );\n\
    if ( MS::kSuccess != stat ) {\n\
        MString str = MString(&quot;Invalid argument 2, usage: threadTestCmd 1 10000&quot;);\n\
        MGlobal::displayError(str);\n\
        return MStatus::kFailure;\n\
    }\n\
\n\
    // start search on an odd number\n\
    if((start % 2) == 0 ) start++;\n\
    // run single threaded\n\
    MTimer timer;\n\
    timer.beginTimer();\n\
    int serialPrimes = SerialPrimes(start, end);\n\
    timer.endTimer();\n\
    double serialTime = timer.elapsedTime();\n\
    // run multithreaded\n\
    timer.beginTimer();\n\
    int parallelPrimes = ParallelPrimes(start, end);\n\
    timer.endTimer();\n\
    double parallelTime = timer.elapsedTime();\n\
    // check for correctness\n\
    if ( serialPrimes != parallelPrimes ) {\n\
        MString str(&quot;Error: Computations inconsistent&quot;);\n\
        MGlobal::displayError(str);\n\
        return MStatus::kFailure;\n\
    }\n\
    // print results\n\
    double ratio = serialTime/parallelTime;\n\
    MString str = MString(&quot;\\nElapsed time for serial computation: &quot;) + serialTime + MString(&quot;s\\n&quot;);\n\
    str += MString(&quot;Elapsed time for parallel computation: &quot;) + parallelTime + MString(&quot;s\\n&quot;);\n\
    str += MString(&quot;Speedup: &quot;) + ratio + MString(&quot;x\\n&quot;);\n\
    MGlobal::displayInfo(str);\n\
    return MStatus::kSuccess;\n\
}\n\
</pre></div>      <div class=\"footer-block\"><a href=\"../html/ac.cmtdialog.htm\" class=\"comments-anchor\" target=\"_blank\"><span class=\"comments-link\">Please send us your comment about this page</span></a></div></div>\n\
   </div></body>\n\
</html>\n\
";